# Copyright 2016-2017 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fnmatch
import struct
import weakref
from xml.etree.ElementTree import fromstring as rfromstring
import zipfile

import six

import pyghmi.constants as pygconst
import pyghmi.exceptions as pygexc
import pyghmi.ipmi.private.session as ipmisession
import pyghmi.ipmi.private.util as util
from pyghmi.ipmi import sdr
import pyghmi.util.webclient as webclient

try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

try:
    range = xrange
except NameError:
    pass


psutypes = {
    0x63: 'CFF v3 550W PT',
    0x64: 'CFF v3 750W PT',
    0x65: 'CFF v3 750W TT',
    0x66: 'CFF v3 1100W PT',
    0x67: 'CFF v3 1600W PT',
    0x68: 'CFF v3 2000W PT',
    0x6B: 'CFF v4 500W PT',
    0x6C: 'CFF v4 750W PT',
    0x6D: 'CFF v4 750W TT',
    0x6E: 'CFF v4 1100W PT',
    0x6F: 'CFF v4 1100W -48Vdc',
    0x70: 'CFF v4 1100W 200-277Vac/240-380Vdc',
    0x71: 'CFF v4 1800W PT',
    0x72: 'CFF v4 2400W PT',
}


def fromstring(inputdata):
    if b'!entity' in inputdata.lower():
        raise Exception('!ENTITY not supported in this interface')
    return rfromstring(inputdata)


def stringtoboolean(originput, name):
    input = originput.lower()
    try:
        num = int(input)
    except ValueError:
        num = None
    if 'enabled'.startswith(input) or 'yes'.startswith(input) or num == 1:
        return True
    elif 'disabled'.startswith(input) or 'no'.startswith(input) or num == 0:
        return False
    raise pygexc.InvalidParameterValue('{0} is an invalid setting for '
                                       '{1}'.format(originput, name))


def fpc_read_ac_input(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x90, data=(1,))
    rsp = rsp['data']
    if len(rsp) == 6:
        rsp = b'\x00' + bytes(rsp)
    return struct.unpack_from('<H', rsp[3:5])[0]


def fpc_read_dc_output(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x90, data=(2,))
    rsp = rsp['data']
    if len(rsp) == 6:
        rsp = b'\x00' + bytes(rsp)
    return struct.unpack_from('<H', rsp[3:5])[0]


def fpc_read_fan_power(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x90, data=(3,))
    rsp = bytes(rsp['data'])
    rsp += b'\x00'
    return struct.unpack_from('<I', rsp[1:])[0] / 100.0


def fpc_read_psu_fan(ipmicmd, number, sz):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa5, data=(number,))
    rsp = bytes(rsp['data'])
    if len(rsp) > 5:
        return struct.unpack_from('<H', rsp[2:4])[0]
    else:
        return struct.unpack_from('<H', rsp[:2])[0]


def fpc_get_psustatus(ipmicmd, number, sz):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x91)
    mask = 1 << (number - 1)
    rsp['data'] = bytearray(rsp['data'])
    if len(rsp['data']) >= 10:
        tmpdata = rsp['data']
        rsp['data'] = list(struct.unpack('<HHHHBB', tmpdata[:10]))
    if len(rsp['data']) == 6:
        statdata = [0]
    else:
        statdata = []
    statdata += rsp['data']
    presence = statdata[3] & mask == mask
    pwrgood = statdata[4] & mask == mask
    throttle = (statdata[6] | statdata[2]) & mask == mask
    health = pygconst.Health.Ok
    states = []
    if presence and not pwrgood:
        health = pygconst.Health.Critical
        states.append('Power input lost')
    if throttle:
        health = pygconst.Health.Critical
        states.append('Throttled')
    if presence:
        states.append('Present')
    else:
        states.append('Absent')
        health = pygconst.Health.Critical
    return (health, states)


def fpc_get_nodeperm(ipmicmd, number, sz):
    try:
        rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa7, data=(number,))
    except pygexc.IpmiException as ie:
        if ie.ipmicode == 0xd5:  # no node present
            return (pygconst.Health.Ok, ['Absent'])
        raise
    health = pygconst.Health.Ok
    states = []
    if len(rsp['data']) == 4:  # different gens handled rc differently
        rsp['data'] = b'\x00' + bytes(rsp['data'])
    elif len(rsp['data']) == 6:  # New FPC format
        rsp['data'] = bytes(rsp['data'][:2]) + bytes(rsp['data'][3:])
    perminfo = bytearray(rsp['data'])[1]
    if sz == 6:  # FPC
        permfail = ('\x02', '\x03')
    else:  # SMM
        permfail = ('\x02',)
    if perminfo & 0x20:
        if rsp['data'][4] in permfail:
            states.append('Insufficient Power')
            health = pygconst.Health.Failed
        elif rsp['data'][3:5] != '\x00\x00':
            states.append('No Power Permission')
            health = pygconst.Health.Failed
    if perminfo & 0x40:
        states.append('Node Fault')
        health = pygconst.Health.Failed
    if rsp['data'][3:5] == '\x00\x00':
        states.append('Absent')
    return (health, states)


def fpc_read_powerbank(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa2)
    return struct.unpack_from('<H', rsp['data'][3:])[0]


def get_psu_count(ipmicmd, variant):
    if variant == 0x26:
        mymsg = ipmicmd.xraw_command(netfn=0x32, command=0xa8)
        builddata = bytearray(mymsg['data'])
        if builddata[13] in (3, 6):
            return 9
        else:
            return 6
    else:
        return variant & 0xf


def fpc_get_dripstatus(ipmicmd, number, sz):
    health = pygconst.Health.Ok
    states = []
    rsp = ipmicmd.xraw_command(0x34, 5)
    rdata = bytearray(rsp['data'])
    number = number - 1
    if rdata[0] & (1 << number) == 0:
        states.append('Absent')
        health = pygconst.Health.Critical
    if rdata[1] & (1 << number) != 0:
        states.append('Leak detected')
        health = pygconst.Health.Critical
    return (health, states)


fpc_sensors = {
    'AC Power': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_ac_input,
    },
    'DC Power': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_dc_output,
    },
    'PSU Power Loss': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_dc_output,
    },
    'Fan Power': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_fan_power
    },
    'PSU Fan Speed': {
        'type': 'Fan',
        'units': 'RPM',
        'provider': fpc_read_psu_fan,
        'elementsfun': get_psu_count,
    },
    'Total Power Capacity': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_powerbank,
    },
    'Node Power Permission': {
        'type': 'Management Subsystem Health',
        'returns': 'tuple',
        'units': None,
        'provider': fpc_get_nodeperm,
        'elements': 2,
    },
    'Power Supply': {
        'type': 'Power Supply',
        'returns': 'tuple',
        'units': None,
        'elements': True,
        'provider': fpc_get_psustatus,
        'elementsfun': get_psu_count,
    },
    'Drip Sensor': {
        'type': 'Management Subsystem Health',
        'elements': True,
        'abselements': 2,
        'returns': 'tuple',
        'units': None,
        'provider': fpc_get_dripstatus,
    }
}


def get_sensor_names(ipmicmd, size):
    global fpc_sensors
    for name in fpc_sensors:
        if size != 6 and name in ('Fan Power', 'Total Power Capacity',
                                  'DC Power'):
            continue
        if size == 6 and name == 'PSU Power Loss':
            continue
        if size != 0x26 and name == 'Drip Sensor':
            continue
        sensor = fpc_sensors[name]

        if 'abselements' in sensor:
            for elemidx in range(sensor['abselements']):
                elemidx += 1
                yield '{0} {1}'.format(name, elemidx)
        elif 'elementsfun' in sensor:
            for elemidx in range(sensor['elementsfun'](ipmicmd, size)):
                elemidx += 1
                yield '{0} {1}'.format(name, elemidx)
        elif 'elements' in sensor:
            for elemidx in range(sensor['elements'] * (size & 0b11111)):
                elemidx += 1
                yield '{0} {1}'.format(name, elemidx)
        else:
            yield name


def get_sensor_descriptions(ipmicmd, size):
    global fpc_sensors
    for name in fpc_sensors:
        if size != 6 and name in ('Fan Power', 'Total Power Capacity',
                                  'DC Power'):
            continue
        if size == 6 and name == 'PSU Power Loss':
            continue
        if size != 0x26 and name == 'Drip Sensor':
            continue
        sensor = fpc_sensors[name]
        if 'abselements' in sensor:
            for elemidx in range(sensor['abselements']):
                elemidx += 1
                yield {'name': '{0} {1}'.format(name, elemidx),
                       'type': sensor['type']}
        elif 'elementsfun' in sensor:
            for elemidx in range(sensor['elementsfun'](ipmicmd, size)):
                elemidx += 1
                yield {'name': '{0} {1}'.format(name, elemidx),
                       'type': sensor['type']}
        elif 'elements' in sensor:
            for elemidx in range(sensor['elements'] * (size & 0b11111)):
                elemidx += 1
                yield {'name': '{0} {1}'.format(name, elemidx),
                       'type': sensor['type']}
        else:
            yield {'name': name, 'type': sensor['type']}


def get_fpc_firmware(bmcver, ipmicmd, fpcorsmm):
    mymsg = ipmicmd.xraw_command(netfn=0x32, command=0xa8)
    builddata = bytearray(mymsg['data'])
    name = None
    if fpcorsmm != 6:  # SMM
        if fpcorsmm >> 5:
            name = 'SMM2'
            buildid = '{0}{1}{2}{3}{4}{5}{6}'.format(
                *[chr(x) for x in builddata[6:13]])
        else:
            name = 'SMM'
            buildid = '{0}{1}{2}{3}{4}{5}{6}'.format(
                *[chr(x) for x in builddata[5:12]])
    elif len(builddata) == 8:
        builddata = builddata[1:]  # discard the 'completion code'
        name = 'FPC'
        buildid = '{0:02X}{1}'.format(builddata[-2], chr(builddata[-1]))
    bmcmajor, bmcminor = [int(x) for x in bmcver.split('.')]
    bmcver = '{0}.{1:02d}'.format(bmcmajor, bmcminor)
    yield (name, {'version': bmcver, 'build': buildid})
    if fpcorsmm == 6:
        yield ('PSOC', {'version': '{0}.{1}'.format(builddata[2],
                                                    builddata[3])})
    else:
        yield ('PSOC', {'version': '{0}.{1}'.format(builddata[3],
                                                    builddata[4])})


def get_sensor_reading(name, ipmicmd, sz):
    value = None
    sensor = None
    health = pygconst.Health.Ok
    states = []
    if name in fpc_sensors and 'elements' not in fpc_sensors[name]:
        sensor = fpc_sensors[name]
        value = sensor['provider'](ipmicmd)
    else:
        bnam, _, idx = name.rpartition(' ')
        idx = int(idx)
        if bnam in fpc_sensors:
            max = -1
            if 'abselements' in fpc_sensors[bnam]:
                max = fpc_sensors[bnam]['abselements']
            elif 'elementsfun' in fpc_sensors[bnam]:
                max = 99
            elif 'elements' in fpc_sensors[bnam]:
                max = fpc_sensors[bnam]['elements'] * sz
            if idx <= max:
                sensor = fpc_sensors[bnam]
                if 'returns' in sensor:
                    health, states = sensor['provider'](ipmicmd, idx, sz)
                else:
                    value = sensor['provider'](ipmicmd, idx, sz)
    if sensor is not None:
        return sdr.SensorReading({'name': name, 'imprecision': None,
                                  'value': value, 'states': states,
                                  'state_ids': [], 'health': health,
                                  'type': sensor['type']},
                                 sensor['units'])
    raise Exception('Sensor not found: ' + name)


class SMMClient(object):

    def __init__(self, ipmicmd, variant):
        self.smm_variant = variant
        self.ipmicmd = weakref.proxy(ipmicmd)
        self.smm = ipmicmd.bmc
        self.username = ipmicmd.ipmi_session.userid
        self.password = ipmicmd.ipmi_session.password
        self._wc = None

    def clear_bmc_configuration(self):
        self.ipmicmd.xraw_command(0x32, 0xad)

    rulemap = {
        'password_reuse_count': 'passwordReuseCheckNum',
        'password_change_interval': 'passwordChangeInterval',
        'password_expiration': 'passwordDurationDays',
        'password_login_failures': 'passwordFailAllowdNum',
        'password_min_length': 'passwordMinLength',
        'password_lockout_period': 'passwordLockoutTimePeriod',
    }

    fanmodes = {
        1: 'Capped_20%',
        2: 'Capped_25%',
        3: 'Capped_30%',
        4: 'Capped_45%',
        0: 'Normal',
        5: 'Boosted',
    }

    def get_bmc_configuration(self):
        settings = {}
        wc = self.wc
        wc.request(
            'POST', '/data',
            ('get=passwordMinLength,passwordForceChange,passwordDurationDays,'
             'passwordExpireWarningDays,passwordChangeInterval,'
             'passwordReuseCheckNum,passwordFailAllowdNum,'
             'passwordLockoutTimePeriod'))
        rsp = wc.getresponse()
        rspbody = rsp.read()
        accountinfo = fromstring(rspbody)
        for rule in self.rulemap:
            ruleinfo = accountinfo.find(self.rulemap[rule])
            if ruleinfo is not None:
                settings[rule] = {'value': int(ruleinfo.text)}
        dwc = self.ipmicmd.xraw_command(0x32, 0x94)
        dwc = bytearray(dwc['data'])
        if len(dwc) not in (3, 4) or dwc[0] == 1:
            rsp = self.ipmicmd.xraw_command(0x34, 3)
            fanmode = self.fanmodes[bytearray(rsp['data'])[0]]
            settings['fanspeed'] = {
                'value': fanmode, 'default': 'Normal',
                'help': ('Adjust the fan speed of the D2 Chassis. Capped '
                         'settings will reduce fan speed for better acoustic '
                         'experience at the expense of performance.  Normal '
                         'is using the Lenovo engineered cooling adjustments '
                         'across the full range. Boosted adds fanspeed to '
                         'the Normal response to provide more aggressive '
                         'cooling.'),
                'possible': [self.fanmodes[x] for x in self.fanmodes]}
        powercfg = self.ipmicmd.xraw_command(0x32, 0xa2)
        powercfg = bytearray(powercfg['data'])
        if len(powercfg) == 5:
            powercfg = powercfg[1:]
        val = powercfg[0]
        if val == 2:
            val = 'N+N'
        elif val == 1:
            val = 'N+1'
        elif val == 0:
            val = 'Disable'
        settings['power_redundancy'] = {
            'default': 'N+1',
            'value': val,
            'possible': ['N+N', 'N+1', 'Disable'],
            'help': ('Configures allowed power budget according to expected '
                     'redundancy. If N+1, power caps will be set to keep '
                     'servers from using more power than the installed power '
                     'supplies could supply if one fails.  If disabled, '
                     'power budget is set to allow nodes to exceed the '
                     'capacity of a single power supply')
        }
        ovs = powercfg[1]
        if ovs == 1:
            ovs = 'Enable'
        elif ovs == 0:
            ovs = 'Disable'
        settings['power_oversubscription'] = {
            'default': 'Enable',
            'value': ovs,
            'possible': ['Enable', 'Disable'],
            'help': ('In redundant power configuration, permit the power '
                     'budget to exceed the capacity of a single power supply '
                     'so long as both power supplies are functioning. This '
                     'excess is limited to an amount that the remaining power '
                     'supply can sustain for a brief period of time in the '
                     'event of losing the other. This excess will be removed '
                     'at the moment a power supply fails so that power '
                     'delivery is at the sustained capacity of the remaining '
                     'supplies.')
        }
        try:
            numbays = 12
            try:
                chassiscapinfo = self.ipmicmd.xraw_command(
                    0x32, 0x9d, data=[13])
            except pygexc.IpmiException as e:
                if e.ipmicode == 201:  # this must be a 2U
                    numbays = 4
                    chassiscapinfo = self.ipmicmd.xraw_command(
                        0x32, 0x9d, data=[5])
                else:
                    raise
            retoffset = 0
            if len(chassiscapinfo['data']) > 10:
                retoffset = 1
            chassiscapstate = self.ipmicmd.xraw_command(
                0x32, 0xa0, data=[numbays + 1])
            capstate = bool(chassiscapstate['data'][retoffset])
            capmin, capmax, protcap, usercap, thermcap = struct.unpack(
                '<HHHHH', chassiscapinfo['data'][retoffset:retoffset + 10])
            settings['chassis_user_cap'] = {
                'value': usercap,
                'help': 'Specify a maximum wattage to consume, this specific '
                        'system implements a range from {0} to {1}.'.format(
                            capmin, capmax)
            }
            settings['chassis_user_cap_active'] = {
                'value': 'Enable' if capstate else 'Disable',
                'help': 'Specify whether the user capping setting should be '
                        'used or not at the chassis level.',
            }
            for baynum in range(numbays):
                baynum += 1
                baycapinfo = self.ipmicmd.xraw_command(
                    0x32, 0x9d, data=[baynum])
                capmin, capmax, protcap, usercap, thermcap = struct.unpack(
                    '<HHHHH', baycapinfo['data'][retoffset:retoffset + 10])
                settings['bay{0}_user_cap'.format(baynum)] = {
                    'value': usercap,
                    'help': 'Specify a maximum wattage for the server in bay '
                            '{0} to consume, this specific system implements '
                            'a range from {1} to {2}'.format(
                                baynum, capmin, capmax)
                }
                settings['bay{0}_protective_cap'.format(baynum)] = {
                    'value': protcap,
                    'help': 'Show the current protective cap for the system '
                            'in bay {0}'.format(baynum)
                }
                baycapstate = self.ipmicmd.xraw_command(
                    0x32, 0xa0, data=[baynum])
                baycapstate = bool(baycapstate['data'][retoffset])
                settings['bay{0}_user_cap_active'.format(baynum)] = {
                    'value': 'Enable' if baycapstate else 'Disable',
                    'help': 'Specify whether the user capping setting should '
                            'be used or not for bay {0}'.format(baynum),
                    'possible': ['Enable', 'Disable'],
                }
        except Exception:
            pass
        try:
            dhcpsendname = self.ipmicmd.xraw_command(0xc, 0x2,
                                                     data=[1, 0xc5, 0, 0])
            dhcpsendname = bytearray(dhcpsendname['data'])
            dhcpsendname = 'Enable' if dhcpsendname[1] == 1 else 'Disable'
            settings['dhcp_sends_hostname'] = {
                'value': dhcpsendname,
                'help': ('Have the device send  hostname as part of its '
                         'DHCP client requests in option 12'),
                'possible': ['Enable', 'Disable']
            }
            dhcpsendvci = self.ipmicmd.xraw_command(0xc, 0x2,
                                                    data=[1, 0xc6, 0, 0])
            dhcpsendvci = bytearray(dhcpsendvci['data'])
            dhcpsendvci = 'Enable' if dhcpsendvci[1] == 1 else 'Disable'
            settings['dhcp_sends_vendor_class_identifier'] = {
                'value': dhcpsendvci,
                'possible': ['Enable', 'Disable'],
                'help': ('Have the device send vendor class identifier '
                         'as part of its DHCP requests in option 60')
            }
        except Exception:
            pass
        try:
            chassisvpd = self.ipmicmd.xraw_command(0x32, 0xb0, data=(5, 0))
            chassisvpd = bytearray(chassisvpd['data'][2:])
            chassisvpd = bytes(chassisvpd).strip()
            if not isinstance(chassisvpd, str):
                chassisvpd = chassisvpd.decode('utf8')
            settings['chassis_model'] = {
                'value': chassisvpd,
                'help': ('Configure the chassis model number')
            }
        except Exception:
            pass
        try:
            chassisvpd = self.ipmicmd.xraw_command(0x32, 0xb0, data=(5, 1))
            chassisvpd = bytearray(chassisvpd['data'][2:])
            chassisvpd = bytes(chassisvpd).strip()
            if not isinstance(chassisvpd, str):
                chassisvpd = chassisvpd.decode('utf8')
            settings['chassis_serial'] = {
                'value': chassisvpd,
                'help': ('Configure the chassis serial number')
            }
        except Exception:
            pass
        return settings

    def set_bay_cap(self, baynum, val):
        payload = struct.pack('<BH', baynum, int(val))
        self.ipmicmd.xraw_command(0x32, 0x9e, data=payload)

    def set_bay_cap_active(self, baynum, val):
        currstate = self.ipmicmd.xraw_command(0x32, 0xa0, data=[baynum])
        currstate = currstate['data']
        if len(currstate) == 5:
            currstate = currstate[1:]
        savemode = currstate[3]
        enable = val.lower().startswith('enable')
        payload = [baynum, 1 if enable else 0, savemode]
        self.ipmicmd.xraw_command(0x32, 0x9f, data=payload)

    def set_bmc_configuration(self, changeset, variant):
        rules = []
        powercfg = [None, None]
        sendhost = None
        sendvci = None
        newserial = None
        newmodel = None
        for key in changeset:
            if not key:
                raise pygexc.InvalidParameterValue('Empty key is invalid')
            if isinstance(changeset[key], six.string_types):
                changeset[key] = {'value': changeset[key]}
            for rule in self.rulemap:
                if fnmatch.fnmatch(rule, key.lower()):
                    rules.append('{0}:{1}'.format(
                        self.rulemap[rule], changeset[key]['value']))
            if fnmatch.fnmatch('power_redundancy', key.lower()):
                if 'n+n'.startswith(changeset[key]['value'].lower()):
                    powercfg[0] = 2
                elif 'n+1'.startswith(changeset[key]['value'].lower()):
                    powercfg[0] = 1
                elif 'disable'.startswith(changeset[key]['value'].lower()):
                    powercfg[0] = 0
            if fnmatch.fnmatch('power_oversubscription', key.lower()):
                ovs = changeset[key]['value']
                ovs = stringtoboolean(changeset[key]['value'],
                                      'power_oversubscription')
                powercfg[1] = 1 if ovs else 0
            if fnmatch.fnmatch('dhcp_sends_hostname', key.lower()):
                sendhost = changeset[key]['value']
                sendhost = stringtoboolean(changeset[key]['value'],
                                           'dhcp_sends_hostname')
            if fnmatch.fnmatch(
                    'dhcp_sends_vendor_class_identifier', key.lower()):
                sendvci = changeset[key]['value']
                sendvci = stringtoboolean(
                    changeset[key]['value'],
                    'dhcp_sends_vendor_class_identifier')
            if fnmatch.fnmatch('chassis_serial', key.lower()):
                newserial = changeset[key]['value']
            if fnmatch.fnmatch('chassis_model', key.lower()):
                newmodel = changeset[key]['value']
            # Variant low 8 bits is the height in U of chassis, so double that
            #  to get maxbays
            numbays = (self.smm_variant & 0x0f) << 1
            for bayn in range(1, numbays + 1):
                if fnmatch.fnmatch('bay{0}_user_cap'.format(bayn),
                                   key.lower()):
                    self.set_bay_cap(bayn, changeset[key]['value'])
                if fnmatch.fnmatch(
                        'bay{0}_user_cap_active'.format(bayn), key.lower()):
                    self.set_bay_cap_active(bayn, changeset[key]['value'])
            if fnmatch.fnmatch('chassis_user_cap', key.lower()):
                self.set_bay_cap(numbays + 1, changeset[key]['value'])
            if fnmatch.fnmatch('chassis_user_cap_active', key.lower()):
                self.set_bay_cap_active(numbays + 1, changeset[key]['value'])
            if fnmatch.fnmatch('fanspeed', key.lower()):
                for mode in self.fanmodes:
                    byteval = mode
                    mode = self.fanmodes[mode]
                    if changeset[key]['value'].lower() == mode.lower():
                        self.ipmicmd.xraw_command(
                            0x32, 0x9b, data=[byteval])
                        break
                else:
                    raise pygexc.InvalidParameterValue(
                        '{0} not a valid mode for fanspeed'.format(
                            changeset[key]['value']))
        if rules:
            rules = 'set={0}'.format(','.join(rules))
            wc = self.wc
            wc.request('POST', '/data', rules)
            wc.getresponse().read()
        if powercfg != [None, None]:
            if variant != 6:
                if None in powercfg:
                    currcfg = self.ipmicmd.xraw_command(0x32, 0xa2)
                    currcfg = bytearray(currcfg['data'])
                    if powercfg[0] is None:
                        powercfg[0] = currcfg[0]
                    if powercfg[1] is None:
                        powercfg[1] = currcfg[1]
                self.ipmicmd.xraw_command(0x32, 0xa3, data=powercfg)
            elif variant == 6:
                if powercfg[0] is not None:
                    self.ipmicmd.xraw_command(0x32, 0xa3, data=powercfg[:1])
                if powercfg[1] is not None:
                    self.ipmicmd.xraw_command(0x32, 0x9c, data=powercfg[1:])
        if sendhost is not None:
            sendhost = 1 if sendhost else 0
            self.ipmicmd.xraw_command(0xc, 1, data=[1, 0xc5, sendhost])
        if sendvci is not None:
            sendvci = 1 if sendvci else 0
            self.ipmicmd.xraw_command(0xc, 1, data=[1, 0xc6, sendvci])
        if newserial:
            newserial = newserial.ljust(10).encode('utf8')
            cmdata = b'\x05\x01' + newserial
            self.ipmicmd.xraw_command(0x32, 0xaf, data=cmdata)
        if newmodel:
            newmodel = newmodel.ljust(10).encode('utf8')
            cmdata = b'\x05\x00' + newmodel
            self.ipmicmd.xraw_command(0x32, 0xaf, data=cmdata)

    def set_user_priv(self, uid, priv):
        if priv.lower() == 'administrator':
            rsp = self.ipmicmd.xraw_command(netfn=6, command=0x46, data=(uid,))
            username = bytes(rsp['data']).rstrip(b'\x00')
            if not isinstance(username, str):
                username = username.decode('utf8')
            wc = self.wc
            wc.request(
                'POST', '/data', 'set=user({0},1,{1},511,,4,15,0)'.format(
                    uid, username))
            rsp = wc.getresponse()
            rsp.read()

    def reseat_bay(self, bay):
        bay = int(bay)
        if bay == -1:
            self.ipmicmd.xraw_command(0x32, 0xf5)
            return
        if bay % 2 == 0:
            # even node may be unable to reseat based on shared io
            try:
                rsp = self.ipmicmd.xraw_command(0x32, 0xc5, data=[1])
            except Exception:
                rsp = {'data': [1]}
            rsp['data'] = bytearray(rsp['data'])
            if rsp['data'][0] == 2:  # shared io
                rsp = self.ipmicmd.xraw_command(0x32, 0xa7, data=[bay - 1])
                rsp['data'] = bytearray(rsp['data'])
                if rsp['data'][1] == 0x80:
                    raise Exception('Unable to reseat bay {0} due to bay {1} '
                                    'being on with shared IO'.format(
                                        bay, bay - 1))
        self.ipmicmd.xraw_command(netfn=0x32, command=0xa4,
                                  data=[bay, 2])

    def get_diagnostic_data(self, savefile, progress=None, autosuffix=False,
                            variant=None):
        if variant == 6:
            raise Exception('Service data not supported on FPC')
        rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0xb1, data=[0])
        if bytearray(rsp['data'])[0] != 0:
            raise Exception("Service data generation already in progress")
        rsp = self.ipmicmd.xraw_command(netfn=0x34, command=0x12, data=[0])
        if bytearray(rsp['data'])[0] != 0:
            raise Exception("Service data generation already in progress")
        rsp['data'] = b'\x01'
        initpct = 1.0
        if progress:
            progress({'phase': 'initializing', 'progress': initpct})
        while bytearray(rsp['data'])[0] != 0:
            ipmisession.Session.pause(3)
            initpct += 3.0
            if initpct > 99.0:
                initpct = 99.0
            rsp = self.ipmicmd.xraw_command(netfn=0x34, command=0x12, data=[1])
            if progress:
                progress({'phase': 'initializing', 'progress': initpct})
        wc = self.wc
        if wc is None:
            raise Exception("Failed to connect to web api")
        if variant and variant >> 5:
            url = '/preview/smm2-ffdc.tgz?ST1={0}'.format(wc.st1)
        else:
            url = '/preview/smm-ffdc.tgz?ST1={0}'.format(wc.st1)
        if autosuffix and not savefile.endswith('.tgz'):
            savefile += '-smm-ffdc.tgz'
        fd = webclient.FileDownloader(wc, url, savefile)
        fd.start()
        while fd.isAlive():
            fd.join(1)
            if progress and wc.get_download_progress():
                progress({'phase': 'download',
                          'progress': 100 * wc.get_download_progress()})
        if progress:
            progress({'phase': 'complete'})
        return savefile

    def process_fru(self, fru):
        smmv1 = self.smm_variant & 0xf0 == 0
        # TODO(jjohnson2): can also get EIOM, SMM, and riser data if warranted
        snum = bytes(self.ipmicmd.xraw_command(
            netfn=0x32, command=0xb0, data=(5, 1))['data'][:])
        mnum = bytes(self.ipmicmd.xraw_command(
            netfn=0x32, command=0xb0, data=(5, 0))['data'][:])
        if not smmv1:
            snum = snum[2:]
            mnum = mnum[2:]
        fru['Serial Number'] = snum.strip(b' \x00\xff').replace(b'\xff', b'')
        fru['Model'] = mnum.strip(b' \x00\xff').replace(b'\xff', b'')
        return fru

    def get_webclient(self):
        cv = self.ipmicmd.certverify
        wc = webclient.SecureHTTPConnection(self.smm, 443, verifycallback=cv)
        wc.vintage = util._monotonic_time()
        wc.connect()
        loginform = urlencode(
            {
                'user': self.username,
                'password': self.password
            }
        )
        wc.request('POST', '/data/login', loginform)
        rsp = wc.getresponse()
        if rsp.status != 200:
            raise Exception(rsp.read())
        authdata = rsp.read()
        authdata = fromstring(authdata)
        for data in authdata.findall('authResult'):
            if int(data.text) != 0:
                raise Exception("Firmware update already in progress")
        for data in authdata.findall('forwardUrl'):
            if 'renew' in data.text:
                raise Exception("Account password has expired on remote "
                                "device")
        wc.st1 = None
        wc.st2 = None
        for data in authdata.findall('st1'):
            wc.st1 = data.text
        for data in authdata.findall('st2'):
            wc.st2 = data.text
        if not wc.st2:
            # This firmware puts tokens in the html file, parse that
            wc.request('GET', '/index.html')
            rsp = wc.getresponse()
            if rsp.status != 200:
                raise Exception(rsp.read())
            indexhtml = rsp.read()
            if not isinstance(indexhtml, str):
                indexhtml = indexhtml.decode('utf8')
            for line in indexhtml.split('\n'):
                if '"ST1"' in line:
                    wc.st1 = line.split()[-1].replace(
                        '"', '').replace(',', '')
                if '"ST2"' in line:
                    wc.st2 = line.split()[-1].replace(
                        '"', '').replace(',', '')
        if not wc.st2:
            wc.request('GET', '/scripts/index.ajs')
            rsp = wc.getresponse()
            body = rsp.read()
            if rsp.status != 200:
                raise Exception(body)
            if not isinstance(body, str):
                body = body.decode('utf8')
            for line in body.split('\n'):
                if '"ST1"' in line:
                    wc.st1 = line.split()[-1].replace(
                        '"', '').replace(',', '')
                if '"ST2"' in line:
                    wc.st2 = line.split()[-1].replace(
                        '"', '').replace(',', '')
        if not wc.st2:
            raise Exception('Unable to locate ST2 token')
        wc.set_header('ST2', wc.st2)
        return wc

    def set_hostname(self, hostname):
        wc = self.wc
        wc.request('POST', '/data', 'set=hostname:' + hostname)
        rsp = wc.getresponse()
        if rsp.status != 200:
            raise Exception(rsp.read())
        rsp.read()
        self.logout()

    def get_hostname(self):
        currinfo = self.get_netinfo()
        self.logout()
        for data in currinfo.find('netConfig').findall('hostname'):
            return data.text

    def get_netinfo(self):
        wc = self.wc
        wc.request('POST', '/data', 'get=hostname')
        rsp = wc.getresponse()
        data = rsp.read()
        if rsp.status == 400:
            wc.request('POST', '/data?get=hostname', '')
            rsp = wc.getresponse()
            data = rsp.read()
        if rsp.status != 200:
            raise Exception(data)
        currinfo = fromstring(data)
        return currinfo

    def set_domain(self, domain):
        wc = self.wc
        wc.request('POST', '/data', 'set=dnsDomain:' + domain)
        rsp = wc.getresponse()
        if rsp.status != 200:
            raise Exception(rsp.read())
        rsp.read()
        self.logout()

    def get_domain(self):
        currinfo = self.get_netinfo()
        self.logout()
        for data in currinfo.find('netConfig').findall('dnsDomain'):
            return data.text

    def get_ntp_enabled(self, variant):
        wc = self.wc
        wc.request('POST', '/data', 'get=ntpOpMode')
        rsp = wc.getresponse()
        info = fromstring(rsp.read())
        self.logout()
        for data in info.findall('ntpOpMode'):
            return data.text == '1'

    def set_ntp_enabled(self, enabled):
        wc = self.wc
        wc.request('POST', '/data', 'set=ntpOpMode:{0}'.format(
            1 if enabled else 0))
        rsp = wc.getresponse()
        result = rsp.read()
        if not isinstance(result, str):
            result = result.decode('utf8')
        self.logout()
        if '<status>ok</status>' not in result:
            raise Exception("Unrecognized result: " + result)

    def set_ntp_server(self, server, index):
        wc = self.wc
        wc.request('POST', '/data', 'set=ntpServer{0}:{1}'.format(
            index + 1, server))
        rsp = wc.getresponse()
        result = rsp.read()
        if not isinstance(result, str):
            result = result.decode('utf8')
        if '<status>ok</status>' not in result:
            raise Exception("Unrecognized result: " + result)
        self.logout()
        return True

    def get_ntp_servers(self):
        wc = self.wc
        wc.request(
            'POST', '/data', 'get=ntpServer1,ntpServer2,ntpServer3')
        rsp = wc.getresponse()
        result = fromstring(rsp.read())
        srvs = []
        for data in result.findall('ntpServer1'):
            srvs.append(data.text)
        for data in result.findall('ntpServer2'):
            srvs.append(data.text)
        for data in result.findall('ntpServer3'):
            srvs.append(data.text)
        self.logout()
        return srvs

    def update_firmware(self, filename, data=None, progress=None, bank=None):
        if progress is None:
            def progress(x):
                return True
        z = None
        if data and hasattr(data, 'read'):
            if zipfile.is_zipfile(data):
                z = zipfile.ZipFile(data)
            else:
                data.seek(0)
        elif data is None and zipfile.is_zipfile(filename):
            z = zipfile.ZipFile(filename)
        if z:
            for tmpname in z.namelist():
                if tmpname.endswith('.rom'):
                    filename = tmpname
                    data = z.open(filename)
                    break
        progress({'phase': 'upload', 'progress': 0.0})
        wc = self.wc
        wc.request('POST', '/data', 'set=fwType:10')  # SMM firmware
        rsp = wc.getresponse()
        rsp.read()
        url = '/fwupload/fwupload.esp?ST1={0}'.format(wc.st1)
        fu = webclient.FileUploader(
            wc, url, filename, data, formname='fileUpload',
            otherfields={'preConfig': 'on'})
        fu.start()
        while fu.isAlive():
            fu.join(3)
            if progress:
                progress({'phase': 'upload',
                          'progress': 100 * wc.get_upload_progress()})
        progress({'phase': 'validating', 'progress': 0.0})
        url = '/data'
        wc.request('POST', url, 'get=fwVersion,spfwInfo')
        rsp = wc.getresponse()
        rsp.read()
        if rsp.status != 200:
            raise Exception('Error validating firmware')
        progress({'phase': 'apply', 'progress': 0.0})
        wc.request('POST', '/data', 'set=securityrollback:1')
        wc.getresponse().read()
        wc.request('POST', '/data', 'set=fwUpdate:1')
        rsp = wc.getresponse()
        rsp.read()
        complete = False
        while not complete:
            ipmisession.Session.pause(3)
            wc.request('POST', '/data', 'get=fwProgress,fwUpdate')
            rsp = wc.getresponse()
            progdata = rsp.read()
            if rsp.status != 200:
                raise Exception('Error applying firmware')
            progdata = fromstring(progdata)
            if progdata.findall('fwUpdate')[0].text == 'invalid signature':
                raise Exception('Firmware signature invalid')
            percent = float(progdata.findall('fwProgress')[0].text)

            progress({'phase': 'apply',
                      'progress': percent})
            complete = percent >= 100.0
        return 'complete'

    def get_inventory_descriptions(self, ipmicmd, variant):
        if variant >> 5 == 0:
            return
        psucount = get_psu_count(ipmicmd, variant)
        for idx in range(psucount):
            yield 'PSU {}'.format(idx + 1)

    def get_inventory_of_component(self, ipmicmd, component):
        psuidx = int(component.replace('PSU ', ''))
        return self.get_psu_info(ipmicmd, psuidx)

    def get_psu_info(self, ipmicmd, psunum):
        psuinfo = ipmicmd.xraw_command(0x34, 0x6, data=(psunum,))
        psuinfo = bytearray(psuinfo['data'])
        psutype = struct.unpack('<H', psuinfo[23:25])[0]
        psui = {}
        if psuinfo[0] != 1:
            return {'Model': 'Unavailable'}
        psui['Revision'] = psuinfo[34]
        psui['Description'] = psutypes.get(
            psutype, 'Unknown ({})'.format(psutype))
        psui['Part Number'] = str(psuinfo[35:47].strip(
            b' \x00\xff').decode('utf8'))
        psui['FRU Number'] = str(psuinfo[47:59].strip(
            b' \x00\xff').decode('utf8'))
        psui['Serial Number'] = str(psuinfo[59:71].strip(
            b' \x00\xff').decode('utf8'))
        psui['Header Code'] = str(psuinfo[71:75].strip(
            b' \x00\xff').decode('utf8'))
        psui['Vendor'] = str(psuinfo[25:29].strip(
            b' \x00\xff').decode('utf8'))
        psui['Manufacturing Date'] = '20{}-W{}'.format(
            psuinfo[77:79].decode('utf8'), psuinfo[75:77].decode('utf8'))
        psui['Primary Firmware Version'] = '{:x}.{:x}'.format(
            psuinfo[80], psuinfo[79])
        psui['Secondary Firmware Version'] = '{:x}.{:x}'.format(
            psuinfo[82], psuinfo[81])
        psui['Model'] = str(psuinfo[5:23].strip(b' \x00\xff').decode('utf8'))
        psui['Manufacturer Location'] = str(
            psuinfo[83:85].strip(b' \x00\xff').decode('utf8'))
        psui['Barcode'] = str(psuinfo[85:108].strip(
            b' \x00\xff').decode('utf8'))
        return psui

    def logout(self):
        wc = self.wc
        wc.request('POST', '/data/logout', None)
        rsp = wc.getresponse()
        rsp.read()
        self._wc = None

    @property
    def wc(self):
        if (not self._wc or self._wc.broken
                or self._wc.vintage < util._monotonic_time() + 30):
            self._wc = self.get_webclient()
        return self._wc



# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import traceback

import pyghmi.ipmi.command as ipmicommand
import pyghmi.ipmi.console as console
import pyghmi.ipmi.private.serversession as serversession
import pyghmi.ipmi.private.session as ipmisession


__author__ = 'jjohnson2@lenovo.com'


class Bmc(serversession.IpmiServer):

    activated = False
    sol = None
    iohandler = None

    def cold_reset(self):
        raise NotImplementedError

    def power_off(self):
        raise NotImplementedError

    def power_on(self):
        raise NotImplementedError

    def power_cycle(self):
        raise NotImplementedError

    def power_reset(self):
        raise NotImplementedError

    def pulse_diag(self):
        raise NotImplementedError

    def power_shutdown(self):
        raise NotImplementedError

    def get_power_state(self):
        raise NotImplementedError

    def is_active(self):
        raise NotImplementedError

    def activate_payload(self, request, session):
        if self.iohandler is None:
            session.send_ipmi_response(code=0x81)
        elif not self.is_active():
            session.send_ipmi_response(code=0x81)
        elif self.activated:
            session.send_ipmi_response(code=0x80)
        else:
            self.activated = True
            solport = list(struct.unpack('BB', struct.pack('!H', self.port)))
            session.send_ipmi_response(
                data=[0, 0, 0, 0, 1, 0, 1, 0] + solport + [0xff, 0xff])
            self.sol = console.ServerConsole(session, self.iohandler)

    def deactivate_payload(self, request, session):
        if self.iohandler is None:
            session.send_ipmi_response(code=0x81)
        elif not self.activated:
            session.send_ipmi_response(code=0x80)
        else:
            session.send_ipmi_response()
            self.sol.close()
            self.activated = False
            self.sol = None

    @staticmethod
    def handle_missing_command(session):
        session.send_ipmi_response(code=0xc1)

    def get_chassis_status(self, session):
        try:
            powerstate = self.get_power_state()
        except NotImplementedError:
            return session.send_ipmi_response(code=0xc1)
        if powerstate in ipmicommand.power_states:
            powerstate = ipmicommand.power_states[powerstate]
        if powerstate not in (0, 1):
            raise Exception('BMC implementation mistake')
        statusdata = [powerstate, 0, 0]
        session.send_ipmi_response(data=statusdata)

    def control_chassis(self, request, session):
        rc = 0
        try:
            directive = request['data'][0]
            if directive == 0:
                rc = self.power_off()
            elif directive == 1:
                rc = self.power_on()
            elif directive == 2:
                rc = self.power_cycle()
            elif directive == 3:
                rc = self.power_reset()
            elif directive == 4:
                # i.e. Pulse a diagnostic interrupt(NMI) directly
                rc = self.pulse_diag()
            elif directive == 5:
                rc = self.power_shutdown()
            if rc is None:
                rc = 0
            session.send_ipmi_response(code=rc)
        except NotImplementedError:
            session.send_ipmi_response(code=0xcc)

    def get_boot_device(self):
        raise NotImplementedError

    def get_system_boot_options(self, request, session):
        if request['data'][0] == 5:  # boot flags
            try:
                bootdevice = self.get_boot_device()
            except NotImplementedError:
                session.send_ipmi_response(data=[1, 5, 0, 0, 0, 0, 0])
            if (type(bootdevice) != int
                    and bootdevice in ipmicommand.boot_devices):
                bootdevice = ipmicommand.boot_devices[bootdevice]
            paramdata = [1, 5, 0b10000000, bootdevice, 0, 0, 0]
            return session.send_ipmi_response(data=paramdata)
        else:
            session.send_ipmi_response(code=0x80)

    def set_boot_device(self, bootdevice):
        raise NotImplementedError

    def set_system_boot_options(self, request, session):
        if request['data'][0] in (0, 3, 4):
            # for now, just smile and nod at boot flag bit clearing
            # implementing it is a burden and implementing it does more to
            # confuse users than serve a useful purpose
            session.send_ipmi_response()
        elif request['data'][0] == 5:
            bootdevice = (request['data'][2] >> 2) & 0b1111
            try:
                bootdevice = ipmicommand.boot_devices[bootdevice]
            except KeyError:
                session.send_ipmi_response(code=0xcc)
                return
            self.set_boot_device(bootdevice)
            session.send_ipmi_response()
        else:
            raise NotImplementedError

    def handle_raw_request(self, request, session):
        try:
            if request['netfn'] == 6:
                if request['command'] == 1:  # get device id
                    return self.send_device_id(session)
                elif request['command'] == 2:  # cold reset
                    return session.send_ipmi_response(code=self.cold_reset())
                elif request['command'] == 0x48:  # activate payload
                    return self.activate_payload(request, session)
                elif request['command'] == 0x49:  # deactivate payload
                    return self.deactivate_payload(request, session)
            elif request['netfn'] == 0:
                if request['command'] == 1:  # get chassis status
                    return self.get_chassis_status(session)
                elif request['command'] == 2:  # chassis control
                    return self.control_chassis(request, session)
                elif request['command'] == 8:  # set boot options
                    return self.set_system_boot_options(request, session)
                elif request['command'] == 9:  # get boot options
                    return self.get_system_boot_options(request, session)
            session.send_ipmi_response(code=0xc1)
        except NotImplementedError:
            session.send_ipmi_response(code=0xc1)
        except Exception:
            session._send_ipmi_net_payload(code=0xff)
            traceback.print_exc()

    @classmethod
    def listen(cls, timeout=30):
        while True:
            ipmisession.Session.wait_for_rsp(timeout)



# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyghmi.ipmi.oem.lenovo import inventory


firmware_fields = (
    inventory.EntryField("Revision", "B"),
    inventory.EntryField("Bios", "16s"),
    inventory.EntryField("Operational ME", "10s"),
    inventory.EntryField("Recovery ME", "10s"),
    inventory.EntryField("RAID 1", "16s"),
    inventory.EntryField("RAID 2", "16s"),
    inventory.EntryField("Mezz 1", "16s"),
    inventory.EntryField("Mezz 2", "16s"),
    inventory.EntryField("BMC", "16s"),
    inventory.EntryField("LEPT", "16s"),
    inventory.EntryField("PSU 1", "16s"),
    inventory.EntryField("PSU 2", "16s"),
    inventory.EntryField("CPLD", "16s"),
    inventory.EntryField("LIND", "16s"),
    inventory.EntryField("WIND", "16s"),
    inventory.EntryField("DIAG", "16s"))

asrock_firmware_fields = (
    inventory.EntryField("Major Firmware Revision", "B"),
    inventory.EntryField("Minor Firmware Revision", "B"),
    inventory.EntryField("Auxiliary Firmware Revision", "4s"))

firmware_cmd = {
    "lenovo": {
        "netfn": 0x06,
        "command": 0x59,
        "data": (0x00, 0xc7, 0x00, 0x00)},
    "asrock": {
        "netfn": 0x3a,
        "command": 0x50,
        "data": (0x02, 0x00, 0x01)},
}

bios_cmd = {
    "lenovo": {
        "netfn": 0x32,
        "command": 0xE8,
        "data": (0x01, 0x01, 0x02)},
    "asrock": {
        "netfn": 0x3a,
        "command": 0x50,
        "data": (0x02, 0x01, 0x01)},
}


def parse_firmware_info(raw, bios_versions=None, asrock=False):
    fields = firmware_fields

    if asrock:
        fields = asrock_firmware_fields

    bytes_read, data = inventory.parse_inventory_category_entry(raw, fields)
    if asrock:
        major_version = data['Major Firmware Revision']
        minor_version = data['Minor Firmware Revision']
        # Asrock RS160 the minor version is Binary Coded Decimal,
        # convert it to Decimal
        minor_version = (0xff & (minor_version >> 4)) * 10 + \
                        (0xf & minor_version)
        aux_reversion = 0
        if str(data['Auxiliary Firmware Revision']) != '':
            aux_reversion = ord(data['Auxiliary Firmware Revision'])

        bmc_version = "%s.%s.%s" % (
            str(major_version),
            str(minor_version),
            str(aux_reversion))

        yield ("BMC", {'version': bmc_version})
        if bios_versions is not None:
            yield ("Bios", {'version': bios_versions[0:]})
    else:
        del data["Revision"]
        for key in data:
            yield (key, {'version': data[key]})
        if bios_versions is not None:
            yield ("Bios_bundle_ver",
                   {'version': bios_versions['new_img_version']})
            yield ("Bios_current_ver",
                   {'version': bios_versions['cur_img_version']})


def parse_bios_number(raw):
    return inventory.parse_bios_number_entry(raw)


def get_categories():
    return {
        "firmware": {
            "idstr": "FW Version",
            "parser": parse_firmware_info,
            "command": firmware_cmd
        },
        "bios_version": {
            "idstr": "Bios Version",
            "parser": parse_bios_number,
            "command": bios_cmd
        }
    }



# Copyright 2015-2017 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import struct
import time
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

import six

import pyghmi.exceptions as exc
import pyghmi.media as media
import pyghmi.redfish.oem.generic as generic
import pyghmi.util.webclient as webclient

hpm_by_filename = {}


class HpmSection(object):
    __slots__ = ['comp_id', 'comp_ver', 'comp_name', 'section_flash', 'data',
                 'hash_size', 'combo_image']


def cstr_to_str(cstr):
    try:
        endidx = cstr.index(b'\x00')
        cstr = cstr[:endidx]
    except Exception:
        pass
    if not isinstance(cstr, str):
        cstr = cstr.decode('utf8')
    return cstr


def read_hpm(filename, data):
    hpminfo = []
    if data:
        hpmfile = data
        hpmfile.seek(0)
    else:
        hpmfile = open(filename, 'rb')
    try:
        hpmfile.seek(0x20)
        skip = struct.unpack('>H', hpmfile.read(2))[0]
        hpmfile.seek(skip + 1, 1)
        sectype, compid = struct.unpack('BB', hpmfile.read(2))
        while sectype == 2:
            currsec = HpmSection()
            currsec.comp_id = compid
            hpmfile.seek(1, 1)
            major, minor, pat = struct.unpack('<BBI', hpmfile.read(6))
            currsec.comp_ver = '{0}.{1}.{2}'.format(major, minor, pat)
            currsec.comp_name = hpmfile.read(21).rstrip(b'\x00')
            currlen = struct.unpack('<I', hpmfile.read(4))[0] - 16
            oemstr = hpmfile.read(4)
            if oemstr != b'OEM\x00':
                raise Exception(
                    'Unrecognized HPM field near {0}'.format(hpmfile.tell()))
            currsec.section_flash = struct.unpack('<I', hpmfile.read(4))[0]
            hashpresent, hdrsize, blocks = struct.unpack('BBB',
                                                         hpmfile.read(3))
            if hashpresent != 1:
                hashpresent = 0
            currsec.hash_size = hashpresent * (256 * blocks + hdrsize)
            hpmfile.seek(5, 1)
            currsec.data = hpmfile.read(currlen)
            hpminfo.append(currsec)
            sectype, compid = struct.unpack('BB', hpmfile.read(2))
        upimg = (hpminfo[1].data[:-hpminfo[1].hash_size]
                 + hpminfo[2].data[:-hpminfo[2].hash_size])
        hpminfo[2].combo_image = upimg
        hpminfo[1].combo_image = upimg
        currpos = hpmfile.tell()
        hpmfile.seek(0, 2)
        endpos = hpmfile.tell()
        if currpos < (endpos - 512):
            raise Exception("Unexpected end of HPM file")
    finally:
        if not data:
            hpmfile.close()
    return hpminfo


class TsmHandler(generic.OEMHandler):
    hostnic = 'usb0'

    def __init__(self, sysinfo, sysurl, webclient, cache=None, fish=None,
                 gpool=None):
        if cache is None:
            cache = {}
        self._wc = None
        self.username = None
        self.password = None
        self.csrftok = None
        self.isipmi = bool(fish)
        self.fish = fish
        self.fishclient = None
        super(TsmHandler, self).__init__(sysinfo, sysurl, webclient, cache,
                                         gpool)
        self.tsm = webclient.thehost
        self._certverify = webclient._certverify

    def get_bmc_configuration(self):
        wc = self.wc
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/dns-info')
        if status != 200:
            raise Exception(repr(rsp))
        settings = {}
        settings['dns_domain'] = {
            'value': rsp['domain_name']
        }
        dnssrvs = []
        for idx in range(3):
            currsrv = rsp.get('dns_server{0}'.format(idx + 1), '::')
            if currsrv and currsrv != '::':
                dnssrvs.append(currsrv)
        settings['dns_servers'] = {'value': ','.join(dnssrvs)}
        rsp, status = wc.grab_json_response_with_status(
            '/api/LockoutPolicystatus')
        if status == 200:
            isenabled = rsp.get('Status', 0) == 1
            if isenabled:
                settings['password_login_failures'] = {'value': rsp.get(
                    'Attemptstimes', 0)}
            else:
                settings['password_login_failures'] = {'value': 0}
        rsp, status = wc.grab_json_response_with_status(
            '/api/GetPWComplex')
        if status == 200:
            settings['password_complexity'] = {'value': rsp.get(
                'pw_complex', 0)}
        return settings

    def set_bmc_configuration(self, changeset):
        dnschgs = {}
        wc = self.wc
        for key in changeset:
            if isinstance(changeset[key], six.string_types):
                changeset[key] = {'value': changeset[key]}
            currval = changeset[key].get('value', None)
            if 'dns_servers'.startswith(key.lower()):
                srvs = currval.split(',')
                for idx in range(3):
                    if idx < len(srvs):
                        dnschgs['dns_server{0}'.format(idx + 1)] = srvs[idx]
                    else:
                        dnschgs['dns_server{0}'.format(idx + 1)] = ''
            if 'dns_domain'.startswith(key.lower()):
                dnschgs['domain_name'] = currval
            if 'password_complexity'.startswith(key.lower()):
                self._set_pass_complexity(currval, wc)
            if 'password_login_failures'.startswith(key.lower()):
                self._set_pass_lockout(currval, wc)
        if dnschgs:
            self._set_dns_config(dnschgs, wc)

    def _set_pass_complexity(self, currval, wc):
        rsp, status = wc.grab_json_response_with_status(
            '/api/SetPWComplex', {'Enable': currval})
        if status != 200:
            raise Exception(repr(rsp))

    def _set_pass_lockout(self, currval, wc):
        rsp, status = wc.grab_json_response_with_status(
            '/api/LockoutPolicystatus')
        if status != 200:
            raise Exception(repr(rsp))
        request = {
            'SameStatus': 0,
            'Lock_min': rsp.get('Locktime', 5),
            'Rest_min': rsp.get('Resettime', 1),
            'Attemptstimes': rsp.get('Attemptstimes', 3)
        }
        if currval == 0:
            request['Enable'] = 0
        else:
            request['Enable'] = 1
            request['Attemptstimes'] = currval
        rsp, status = wc.grab_json_response_with_status(
            '/api/SetLockoutPolicy', request)
        if status != 200:
            raise Exception(repr(rsp))

    def _set_dns_config(self, dnschgs, wc):
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/dns-info')
        if status != 200:
            raise Exception(repr(rsp))
        rsp['domain_manual'] = 1
        for i in range(3):
            keyn = 'dns_server{0}'.format(i + 1)
            if rsp[keyn] == '::':
                rsp[keyn] = ''
        for chg in dnschgs:
            rsp[chg] = dnschgs[chg]
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/dns-info', rsp, method='PUT')
        if status != 200:
            raise Exception(repr(rsp))
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/dns/restart', {'dns_status': 1}, method='PUT')
        if status != 200:
            raise Exception(repr(rsp))

    def clear_uefi_configuration(self):
        if not self.fishclient:
            self.init_redfish()
        return self.fishclient.clear_system_configuration()

    def get_uefi_configuration(self, hideadvanced=True):
        if not self.fishclient:
            self.init_redfish()
        return self.fishclient.get_system_configuration(hideadvanced)

    def set_uefi_configuration(self, changeset):
        if not self.fishclient:
            self.init_redfish()
        return self.fishclient.set_system_configuration(changeset)

    def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
        wc = self.wc
        wc.grab_json_response('/api/mini_ffdc', {'action': 'trigger'})
        status = 1
        percent = 0
        while status == 1:
            time.sleep(5)
            check = wc.grab_json_response('/api/mini_ffdc',
                                          {'action': 'check'})
            status = check.get('status', -1)
            if progress:
                progress({'phase': 'initializing', 'progress': float(percent)})
            percent += 1
        if status != 2:
            raise Exception(
                "Unknown error generating service data: " + repr(check))
        if autosuffix and not savefile.endswith('.tar'):
            savefile += '.tar'
        fd = webclient.FileDownloader(wc, '/api/mini_ffdc/package', savefile)
        fd.start()
        while fd.isAlive():
            fd.join(1)
            if progress:
                currprog = self.wc.get_download_progress()
                if currprog:
                    progress({'phase': 'download',
                              'progress': 100 * currprog})
        if fd.exc:
            raise fd.exc
        if progress:
            progress({'phase': 'complete'})
        return savefile

    def init_redfish(self):
        self.fishclient = self.fish.Command(
            self.tsm, self.username, self.password,
            verifycallback=self._certverify)

    def get_ntp_enabled(self):
        wc = self.wc
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/date-time')
        if status != 200:
            raise Exception(repr(rsp))
        return rsp.get('ntp_auto_date', 0) > 0

    def set_ntp_enabled(self, enabled):
        wc = self.wc
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/date-time')
        if status != 200:
            raise Exception(repr(rsp))
        rsp['ntp_auto_date'] = 1 if enabled else 0
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/date-time', rsp, method='PUT')
        if status != 200:
            raise Exception(repr(rsp))

    def get_ntp_servers(self):
        wc = self.wc
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/date-time')
        if status != 200:
            raise Exception(repr(rsp))
        srvs = []
        pntp = rsp.get('primary_ntp', None)
        if pntp:
            srvs.append(pntp)
        pntp = rsp.get('secondary_ntp', None)
        if pntp:
            srvs.append(pntp)
        return srvs

    def set_ntp_server(self, server, index=0):
        wc = self.wc
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/date-time')
        if status != 200:
            raise Exception(repr(rsp))
        if index == 0:
            rsp['primary_ntp'] = server
        elif index == 1:
            rsp['secondary_ntp'] = server
        rsp['ntp_auto_date'] = 1
        rsp, status = wc.grab_json_response_with_status(
            '/api/settings/date-time', rsp, method='PUT')
        if status != 200:
            raise Exception(repr(rsp))

    def get_firmware_inventory(self, components, raisebypass=True,
                               ipmicmd=None):
        wc = self.wc
        fwinf, status = wc.grab_json_response_with_status(
            '/api/DeviceVersion')
        gotinfo = False
        if status < 200 or status >= 300:
            raise Exception('Error connecting to HTTP API')
        for biosinf in fwinf:
            if biosinf.get('device', None) != 1:
                continue
            if not biosinf.get('buildname', False):
                break
            biosres = {
                'build': biosinf['buildname']
            }
            if biosinf.get('main', False):
                biosres['version'] = '{0}.{1}'.format(
                    biosinf['main'][0], biosinf['main'][1:])
            yield ('UEFI', biosres)
            gotinfo = True
            break
        for lxpminf in fwinf:
            if lxpminf.get('device', None) != 2:
                continue
            if not lxpminf.get('buildname', False):
                break
            lxpmres = {
                'build': lxpminf['buildname']
            }
            if lxpminf.get('main', False):
                subver = lxpminf.get('sub', 0)
                lxpmres['version'] = '{0}.{1:02x}'.format(
                    lxpminf['main'], subver)
            yield ('LXPM', lxpmres)
        if ipmicmd:
            rsp = ipmicmd.xraw_command(0x3c, 0x40, data=(7, 2))
            buildid = cstr_to_str(bytes(rsp['data']))
            yield ('LXPM Windows Driver Bundle', {'build': buildid})
            rsp = ipmicmd.xraw_command(0x3c, 0x40, data=(7, 3))
            buildid = cstr_to_str(bytes(rsp['data']))
            yield ('LXPM Linux Driver Bundle', {'build': buildid})
        name = 'TSM'
        fwinf, status = wc.grab_json_response_with_status('/api/get-sysfwinfo')
        if status != 200:
            raise Exception('Error {0} retrieving TSM version: {1}'.format(
                status, fwinf))
        for cinf in fwinf:
            if 'fw_ver' not in cinf:
                continue
            if cinf.get('buildname', None) == 'N/A':
                continue
            if '.' not in cinf['fw_ver']:
                continue
            bmcinf = {
                'version': cinf['fw_ver'],
                'build': cinf['buildname'],
                'date': cinf['builddate'],
            }
            yield (name, bmcinf)
            gotinfo = True
            name += ' Backup'
        if not gotinfo:
            raise Exception("Unable to retrieve firmware information")
        if raisebypass:
            raise exc.BypassGenericBehavior()

    @property
    def wc(self):
        self.fwid = None
        if self._wc:
            rsp, status = self._wc.grab_json_response_with_status(
                '/api/chassis-status')
            if status == 200:
                return self._wc
        authdata = {
            'username': self.username,
            'password': self.password,
        }
        wc = webclient.SecureHTTPConnection(self.tsm, 443,
                                            verifycallback=self._certverify,
                                            timeout=180)
        wc.set_header('Content-Type', 'application/json')
        rsp, status = wc.grab_json_response_with_status('/api/session',
                                                        authdata)
        if status == 403:
            wc.set_header('Content-Type', 'application/x-www-form-urlencoded')
            rsp, status = wc.grab_json_response_with_status(
                '/api/session', urlencode(authdata))

        if status < 200 or status >= 300:
            raise Exception('Error establishing web session')
        self.csrftok = rsp['CSRFToken']
        wc.set_header('X-CSRFTOKEN', self.csrftok)
        self._wc = wc
        return wc

    def update_firmware(self, filename, data=None, progress=None, bank=None):
        wc = self.wc
        wc.set_header('Content-Type', 'application/json')
        basefilename = os.path.basename(filename)
        if filename.endswith('.hpm'):
            return self.update_hpm_firmware(filename, progress, wc, data)
        elif 'uefi' in basefilename and filename.endswith('.rom'):
            return self.update_sys_firmware(filename, progress, wc, data=data)
        elif 'amd-sas' in basefilename and filename.endswith('.bin'):
            return self.update_sys_firmware(filename, progress, wc, data=data,
                                            type='bp')
        elif (('lxpm' in basefilename or 'fw_drv' in basefilename)
                and filename.endswith('.img')):
            return self.update_lxpm_firmware(filename, progress, wc, data)
        else:
            raise Exception('Unsupported filename {0}'.format(filename))

    def update_lxpm_firmware(self, filename, progress, wc, data):
        hdrs = wc.stdheaders.copy()
        hdrs['Content-Length'] = 0
        rsp = wc.grab_json_response_with_status(
            '/api/maintenance/LXPMUploadMode',
            method='PUT', headers=hdrs)
        # name fwimage filname filename application/x-raw-disk-image...
        fu = webclient.FileUploader(
            wc, '/api/maintenance/LXPMUpload',
            filename, data, formname='fwimage')
        fu.start()
        while fu.isAlive():
            fu.join(3)
            if progress:
                progress({
                    'phase': 'upload',
                    'progress': 100 * wc.get_upload_progress()})
        if progress:
            progress({
                'phase': 'apply',
                'progress': 0.0}
            )
        wc.grab_json_response('/api/maintenance/LXPMImageSplit', {'type': 3})
        completion = False
        while not completion:
            rsp = wc.grab_json_response('/api/maintenance/LXPMstatus')
            if rsp.get('state') == 0 and rsp.get('progress') == 4:
                break
        wc.grab_json_response_with_status(
            '/api/maintenance/Outofflash', method='PUT', headers=hdrs)
        return 'complete'

    def update_sys_firmware(self, filename, progress, wc, type='uefi',
                            data=None):
        if type == 'bp':
            rsp = wc.grab_json_response_with_status('/api/chassis-status')
            if rsp[0]['power_status'] == 1:
                raise Exception("Cannot update BP firmware while system is on")
            updatemode = 'BPUploadMode'
            fileupload = 'BPfileUpload'
            startit = 'BPUpgradeStart'
            statusname = 'BPstatus'
        else:
            updatemode = 'flash'
            rsp = wc.grab_json_response_with_status(
                '/api/maintenance/BIOSremoteSave',
                {"tftpip": "",
                 "tftpfile": ""}
            )
            fileupload = 'firmware/BIOS'
            startit = 'BIOSstart'
            statusname = 'BIOSstatus'
        hdrs = wc.stdheaders.copy()
        hdrs['Content-Length'] = 0
        rsp = wc.grab_json_response_with_status(
            '/api/maintenance/{0}'.format(updatemode),
            method='PUT', headers=hdrs)
        fu = webclient.FileUploader(
            wc, '/api/maintenance/{0}'.format(fileupload), filename, data,
            formname='fwimage')
        fu.start()
        while fu.isAlive():
            fu.join(3)
            if progress:
                progress({
                    'phase': 'upload',
                    'progress': 100 * wc.get_upload_progress()})
        if progress:
            progress({
                'phase': 'apply',
                'progress': 0.0}
            )
        rsp = wc.grab_json_response_with_status(
            '/api/maintenance/{0}'.format(startit))
        applypct = 0.0
        if rsp[1] >= 200 and rsp[1] < 300 and rsp[0]['wRet'] == 0:
            updone = False
            while not updone:
                rsp = wc.grab_json_response(
                    '/api/maintenance/{0}'.format(statusname))
                if rsp.get('state', 0) == 9:
                    break
                if rsp.get('state', 0) in (6, 10):
                    raise Exception('Update Failure')
                if (rsp.get('state', 0) == 8
                        and rsp.get('progress', 0) > 0 and progress):
                    progress({
                        'phase': 'apply',
                        'progress': 70 + float(rsp.get(
                            'progress', 0)) / 100 * 30})
                elif type == 'bp' and rsp.get('state', 0) == 1:
                    break
                elif progress and applypct < 70:
                    applypct += 1.4
                    progress({'phase': 'apply', 'progress': applypct})
            if type == 'bp':
                rsp = wc.grab_json_response('/api/maintenance/BPfinish')
                hdrs = wc.stdheaders.copy()
                hdrs['Content-Length'] = 0
                rsp = wc.grab_json_response_with_status(
                    '/api/maintenance/Outofflash', method='PUT', headers=hdrs)
                return 'complete'
            return 'pending'
        raise Exception('Update Failure')

    def update_hpm_firmware(self, filename, progress, wc, data):
        rsp = wc.grab_json_response('/api/maintenance/hpm/freemem')
        if 'MemFree' not in rsp:
            raise Exception('System Not Ready for update')
        if filename not in hpm_by_filename:
            hpminfo = read_hpm(filename, data)
            if len(hpminfo) != 3:
                raise Exception(
                    'This HPM update is currently not supported')
            hpm_by_filename[filename] = read_hpm(filename, data)
        else:
            hpminfo = hpm_by_filename[filename]
        rsp, status = wc.grab_json_response_with_status(
            '/api/maintenance/hpm/updatemode', method='PUT')
        # first segment, make sure it is mmc,
        # then do the preparecomponents with the following payload
        if status != 200:
            raise Exception(rsp)
        uid = rsp['unique_id']
        self.fwid = uid
        payload = {
            'FWUPDATEID': uid,
            'COMPONENT_ID': 1,
            'COMPONENT_DATA_LEN': len(hpminfo[0].data),
            'IS_MMC': 1,
        }
        rsp, status = wc.grab_json_response_with_status(
            '/api/maintenance/hpm/preparecomponents', payload, method='PUT')
        if status < 200 or status >= 300:
            wc.grab_json_response_with_status(
                '/api/maintenance/hpm/exitupdatemode', {'FWUPDATEID': uid},
                method='PUT')
            raise Exception(rsp)
        fu = webclient.FileUploader(
            wc, '/api/maintenance/hpm/mmcfw', 'blob', hpminfo[0].data, 'mmc')
        if progress:
            progress({'phase': 'upload', 'progress': 0.0})
        fu.start()
        while fu.isAlive():
            fu.join(3)
            if progress:
                progress({
                    'phase': 'upload',
                    'progress': 50 * wc.get_upload_progress()})
        del payload['IS_MMC']
        payload['SECTION_FLASH'] = hpminfo[0].section_flash
        rsp, status = wc.grab_json_response_with_status(
            '/api/maintenance/hpm/flash', payload, method='PUT')
        percent = 0
        while percent < 100:
            rsp, status = wc.grab_json_response_with_status(
                '/api/maintenance/hpm/upgradestatus?COMPONENT_ID=1')
            if status < 200 or status >= 300:
                raise Exception(rsp)
            percent = rsp['PROGRESS']
            if progress:
                progress({
                    'phase': 'apply',
                    'progress': .5 * percent})
            if percent < 100:
                time.sleep(3)
        if progress:
            progress({'phase': 'validating', 'progress': 0.0})
        del payload['SECTION_FLASH']
        rsp, status = wc.grab_json_response_with_status(
            '/api/maintenance/hpm/verifyimage', payload, method='PUT')
        percent = 0
        while percent < 100:
            rsp, status = wc.grab_json_response_with_status(
                '/api/maintenance/hpm/verifyimagestatus?COMPONENT_ID=1')
            if status < 200 or status >= 300:
                raise Exception(rsp)
            percent = rsp['PROGRESS']
            if progress:
                progress({
                    'phase': 'validating',
                    'progress': 0.5 * percent})
            if percent < 100:
                time.sleep(3)
        rsp, status = wc.grab_json_response_with_status(
            '/api/maintenance/hpm/exitupdatemode', {'FWUPDATEID': uid},
            method='PUT')
        fu = webclient.FileUploader(wc, '/api/maintenance/firmware/firmware',
                                    'blob', hpminfo[1].combo_image, 'fwimage')
        fu.start()
        while fu.isAlive():
            fu.join(3)
            if progress:
                progress({
                    'phase': 'upload',
                    'progress': 50 * wc.get_upload_progress() + 50})
        rsp = wc.grab_json_response('/api/maintenance/firmware/verification')
        upgradeparms = {
            'preserve_config': 1,
            'flash_status': 1,
        }
        rsp, status = wc.grab_json_response_with_status(
            '/api/maintenance/firmware/upgrade',
            upgradeparms, method='PUT')
        if progress:
            progress({'phase': 'apply', 'progress': 50.0})
        applied = False
        while not applied:
            rsp = wc.grab_json_response(
                '/api/maintenance/firmware/flash-progress')
            percent = float(rsp['progress'].split('%')[0])
            percent = percent * 0.5 + 50
            if progress:
                progress({'phase': 'apply', 'progress': percent})
            if rsp['progress'] == '100% done' and rsp['state'] == 0:
                applied = True
                break
            time.sleep(3)
        hdrs = wc.stdheaders.copy()
        hdrs['Content-Length'] = 0
        rsp = wc.grab_json_response_with_status(
            '/api/maintenance/reset', method='POST', headers=hdrs)
        self._wc = None
        return 'complete'

    def _detach_all_media(self, wc, slots):
        for slot in slots:  # Stop all active redirections to reconfigure
            if slot['redirection_status'] != 0:
                wc.grab_json_response(
                    '/api/settings/media/remote/stop-media',
                    {'image_name': slot['image_name'],
                     'image_type': slot['media_type'],
                     'image_index': slot['media_index']})

    def detach_remote_media(self):
        wc = self.wc
        slots = wc.grab_json_response(
            '/api/settings/media/remote/configurations')
        self._detach_all_media(wc, slots)
        if not self.isipmi:
            raise exc.BypassGenericBehavior()

    def _allocate_slot(self, slots, filetype, wc, server, path):
        currhdds = []
        currisos = []
        for slot in slots:
            if slot['image_name']:
                if slot['media_type'] == 1:
                    currisos.append(slot['image_name'])
                elif slot['media_type'] == 4:
                    currhdds.append(slot['image_name'])
                else:
                    raise exc.UnsupportedFunctionality(
                        'Unrecognized mounted image: ' + repr(slot))
        hddslots = len(currhdds)
        cdslots = len(currisos)
        if filetype == 1:
            cdslots += 1
        elif filetype == 4:
            hddslots += 1
        else:
            raise exc.UnsupportedFunctionality('Unknown slot type requested')
        gensettings = wc.grab_json_response('/api/settings/media/general')
        samesettings = gensettings['same_settings'] == 1
        if samesettings:
            hds = gensettings['cd_remote_server_address']
            hdp = gensettings['cd_remote_source_path'].replace('\\/', '/')
        else:
            hds = gensettings['hd_remote_server_address']
            hdp = gensettings['hd_remote_source_path'].replace('\\/', '/')
        if filetype == 1 and (currisos or (samesettings and currhdds)):
            if gensettings['cd_remote_server_address'] != server:
                raise exc.UnsupportedFunctionality(
                    'Cannot mount ISO images from multiple '
                    'servers at a time')
            if gensettings['cd_remote_source_path'].replace(
                    '\\/', '/') != path:
                raise exc.UnsupportedFunctionality(
                    'Cannot mount ISO images from different '
                    'directories at a time')
        if filetype == 4 and currhdds:
            if hds != server:
                raise exc.UnsupportedFunctionality(
                    'Cannot mount IMG images from multiple servers at a time')
            if hdp != path:
                raise exc.UnsupportedFunctionality(
                    'Cannot mount IMG images from muliple directories at a '
                    'time')
        self._detach_all_media(wc, slots)
        if filetype == 1 or (samesettings and currhdds):
            gensettings['cd_remote_server_address'] = server
            gensettings['cd_remote_source_path'] = path
            gensettings['cd_remote_share_type'] = 'nfs'
            gensettings['mount_cd'] = 1
        elif filetype == 4:
            gensettings['same_settings'] = 0
            gensettings['hd_remote_server_address'] = server
            gensettings['hd_remote_source_path'] = path
            gensettings['hd_remote_share_type'] = 'nfs'
            gensettings['mount_hd'] = 1
        gensettings['remote_media_support'] = 1
        gensettings['cd_remote_password'] = ''
        gensettings['hd_remote_password'] = ''
        wc.grab_json_response_with_status('/api/settings/media/general',
                                          gensettings, method='PUT')
        # need to calibrate instances correctly
        currinfo, status = wc.grab_json_response_with_status(
            '/api/settings/media/instance')
        currinfo['num_cd'] = cdslots
        currinfo['num_hd'] = hddslots
        if currinfo['kvm_num_cd'] > cdslots:
            currinfo['kvm_num_cd'] = cdslots
        if currinfo['kvm_num_hd'] > hddslots:
            currinfo['kvm_num_hd'] = hddslots
        wc.grab_json_response_with_status(
            '/api/settings/media/instance', currinfo, method='PUT')
        images = wc.grab_json_response('/api/settings/media/remote/images')
        tries = 20
        while tries and not images:
            tries -= 1
            time.sleep(1)
            images = wc.grab_json_response('/api/settings/media/remote/images')
        for iso in currisos:
            self._exec_mount(iso, images, wc)
        for iso in currhdds:
            self._exec_mount(iso, images, wc)

    def _exec_mount(self, name, images, wc):
        for img in images:
            if img['image_name'] == name:
                break
        else:
            raise exc.InvalidParameterValue(
                'Unable to locate image {0}'.format(name))
        wc.grab_json_response(
            '/api/settings/media/remote/start-media',
            {'image_name': name, 'image_type': img['image_type'],
             'image_index': img['image_index']})

    def upload_media(self, filename, progress=None, data=None):
        raise exc.UnsupportedFunctionality(
            'Remote media upload not supported on this system')

    def list_media(self, fishclient=None):
        wc = self.wc
        rsp = wc.grab_json_response('/api/settings/media/general')
        cds = rsp['cd_remote_server_address']
        cdpath = rsp['cd_remote_source_path']
        cdproto = rsp['cd_remote_share_type']
        if rsp['same_settings'] == 1:
            hds = cds
            hdpath = cdpath
            hdproto = cdproto
        else:
            hds = rsp['hd_remote_server_address']
            hdpath = rsp['hd_remote_source_path']
            hdproto = rsp['hd_remote_share_type']
        slots = wc.grab_json_response(
            '/api/settings/media/remote/configurations')
        for slot in slots:
            if slot['redirection_status'] == 1:
                url = None
                if slot['media_type'] == 1:
                    url = '{0}://{1}{2}'.format(
                        cdproto, cds, cdpath)
                elif slot['media_type'] == 4:
                    url = '{0}://{1}{2}'.format(
                        hdproto, hds, hdpath)
                if url:
                    yield media.Media(slot['image_name'], url)

    def attach_remote_media(self, url, user, password, vmurls):
        if not url.startswith('nfs://'):
            raise exc.UnsupportedFunctionality(
                'Only nfs:// urls are supported by this system')
        path = url.replace('nfs://', '')
        server, path = path.split('/', 1)
        path, filename = path.rsplit('/', 1)
        path = '/' + path
        filetype = filename.rsplit('.')[-1]
        if filetype == 'iso':
            filetype = 1
        elif filetype == 'img':
            filetype = 4
        else:
            raise exc.UnsupportedFunctionality(
                'Only iso and img files supported')
        wc = self.wc
        mountslots = wc.grab_json_response(
            '/api/settings/media/remote/configurations')
        images = wc.grab_json_response('/api/settings/media/remote/images')
        currtypeenabled = False
        for slot in mountslots:
            if slot['image_name'] == filename:
                return  # Already mounted...
        for img in images:
            if img['image_name'] == filename:
                break
            if img['image_type'] == filetype:
                currtypeenabled = True
        else:
            if currtypeenabled:
                raise exc.UnsupportedFunctionality(
                    'This system cannot mount images '
                    'from different locations at the same time')
            img = None
        for slot in mountslots:
            if slot['media_type'] != filetype:
                continue
            if slot['redirection_status'] == 0:
                break
        else:
            self._allocate_slot(mountslots, filetype, wc, server, path)
        images = wc.grab_json_response('/api/settings/media/remote/images')
        self._exec_mount(filename, images, wc)
        if not self.isipmi:
            raise exc.BypassGenericBehavior()
